# -*- coding: utf-8 -*-
"""
Created on Wed Aug 30 19:42:37 2017
@author: Alvaro Magdaleno Gonzalez

Modified on Jun 2021 by Jose Manuel Martin Cabeza
"""
from StruPy2D.Material import defMat
from StruPy2D.Section import defSec
from StruPy2D.Load import LoadList

import numpy as np
from math import sin, cos, pi, atan
import contextlib as ctxlb

from StruPy2D.config import TOL

@ctxlb.contextmanager
def printoptions(*args, **kwargs):
    original = np.get_printoptions()
    np.set_printoptions(*args, **kwargs)
    try:
        yield
    finally: 
        np.set_printoptions(**original)
        
class Bar(object):
    def __init__(self,ID,p1,p2,mat=defMat,sec=defSec):
        if(type(p1).__name__ != "Node" or type(p2).__name__ != "Node"):
            raise TypeError("Bar nodes must be objects of Class Node.")
            
        if(type(mat).__name__ != "Material"):
            raise TypeError("Material must be an object of Class Material.")
            
        if(type(sec).__name__ != "Section"):
            raise TypeError("Section must be an object of Class Section.")
            
        if(np.abs(p1.dist(p2)) < TOL):
            raise ValueError("Points must be separated at least "+str(TOL)+" units.")
        
        self.p1 = p1
        self.p2 = p2
        p1.setBar(self)
        p2.setBar(self)
        self.nBu = 1 #Propiedad del axil medio en la barra para caso de pandeo global
        
        
        self.L = p1.dist(p2) #Longitud y angulo de la barra
        if(p1.x == p2.x):
            self.th = pi/2
            if(p1.y > p2.y):
                self.th += pi
        else:        
            self.th = atan((p2.y - p1.y)/(p2.x - p1.x))
            if(p1.x > p2.x):
                self.th += pi
        
        L = np.array([[cos(self.th), sin(self.th), 0],[-sin(self.th), cos(self.th), 0],[0, 0, 1]])
        self.T = np.append(np.append(L,np.zeros((3,3)),1),np.append(np.zeros((3,3)),L,1),0)
        
        self.mat = mat
        self.sec = sec
        
        try:
            self.ID = str(ID)
        except ValueError as err:
            err.args = ['ID must be convertible to string.']
            raise err
            
        self.releases = [False]*6 # lista de GDL liberados de la barra
        self.relF = [0]*6 # Lista de las fuerzas aplicadas en los gdl liberados
        self.getGlobal() # Computes local and global matrices
        self.loads = LoadList() # Lista de cargas que sufre la barra
        self.Femp = np.array(np.zeros((6,1)))
        self.FempR = self.Femp
        self.dupli = False
        self._weight = None # Objecto carga DIS que representa su peso propio
        
    def __str__(self):
        print('Bar {}, from ({}, {}) to ({}, {})'.format(self.ID,self.p1.x,self.p1.y,self.p2.x,self.p2.y))
        print('Length = {}'.format(self.L))
        print('Angle (deg) = {}'.format(self.th*180/pi))
        print('Material:  ',self.mat.ID,sep='')
        print('Section:   ',self.sec.ID,sep='')
        
        if(sum(self.releases) > 0):
            string = 'Releases at DOFs:'
            for i in range(6):
                if self.releases[i]:
                    string = string + ' ' + str(i)
            print(string)
            
        return ''
    
    def __repr__(self):
        return str(self.ID)
    
    def __eq__(self, key):
        if(type(key).__name__ == 'Bar'):
            if(key.p1 == self.p1 and key.p2 == self.p2):
                return True
            else:
                return False
        else:
            raise ValueError("Bar objects are only comparable to other Bar objects")
    
    def __ne__(self, key):
        if(type(key).__name__ == 'Bar'):
            if(key.p1 != self.p1 or key.p2 != self.p2):
                return True
            else:
                return False
        else:
            raise ValueError("Bar objects are only comparable to other Bar objects")
        
    def setMat(self, mat):
        if(type(mat).__name__ != "Material"):
            raise TypeError("Material must be an object of Class Material.")
            
        self.mat = mat
        self.getGlobal()
        self.getFemp()
        
    def setSec(self, sec):
        if(type(sec).__name__ != "Section"):
            raise TypeError("Section must be an object of Class Section.")
            
        self.sec = sec
        self.getGlobal()
        self.getFemp()
        
    def seeKe(self, precision = 3):
        with printoptions(precision=precision, suppress=True):
            print(self.Ke)
    
    def seeNKGe(self, precision = 3):
        with printoptions(precision=precision, suppress=True):
            print(self.NKGe)
            
    def seeMe(self, precision = 3):
        with printoptions(precision=precision, suppress=True):
            print(self.Me)
        
    def updateLth(self):
        """Actualiza la longitud y ángulo de la barra por un posible cambio de
        la posición de algún nodo"""
        p1 = self.p1
        p2 = self.p2
        
        self.L = p1.dist(p2)
        if(p1.x == p2.x):
            self.th = pi/2
            if(p1.y > p2.y):
                self.th += pi
        else:        
            self.th = atan((p2.y - p1.y)/(p2.x - p1.x))
            if(p1.x > p2.x):
                self.th += pi
            
        L = np.array([[cos(self.th), sin(self.th), 0],[-sin(self.th), cos(self.th), 0],[0, 0, 1]])
        self.T = np.append(np.append(L,np.zeros((3,3)),1),np.append(np.zeros((3,3)),L,1),0)
        
        self.getGlobal()
        self.getFemp()
        
    def getKe(self):
        al = self.sec.A*self.mat.E/self.L
        be = self.mat.E*self.sec.I/self.L**3
        L = self.L
        
        Ke = np.array([[al,  0,      0,         -al,  0,      0],
                       [0,   12*be,  6*be*L,    0,   -12*be,  6*be*L],
                       [0,   6*be*L, 4*be*L**2, 0,   -6*be*L, 2*be*L**2],
                       [-al, 0,      0,         al,  0,       0],
                       [0,   -12*be, -6*be*L,   0,   12*be,  -6*be*L],
                       [0,   6*be*L, 2*be*L**2, 0,   -6*be*L, 4*be*L**2]])
        
        self.Ke = np.mat(Ke)
        
        
    def getNKGe(self):
        L = self.L
        
        NKGe = self.nBu*np.array([np.zeros(6),
                 [0, 6/(5*L), 1/10, 0, -6/(5*L), 1/10],
                 [0, 1/10, 2*L/15, 0, -1/10, -L/30],
                 np.zeros(6),
                 [0, -6/(5*L), -1/10, 0, 6/(5*L), -1/10],
                 [0, 1/10, -L/30, 0, -1/10, 2*L/15]])

        self.NKGe = np.mat(NKGe)    
        
        
    def getMe(self):
        fac = self.mat.rho*self.sec.A*self.L/420;
        L = self.L
        
        Me = fac*np.array([[140, 0,     0,       70,  0,     0],
                           [0,   156,   22*L,    0,   54,    -13*L],
                           [0,   22*L,  4*L**2,  0,   13*L,  -3*L**2],
                           [70,  0,     0,       140, 0,     0],
                           [0,   54,    13*L,    0,   156,   -22*L],
                           [0,   -13*L, -3*L**2, 0,   -22*L, 4*L**2]])
    
        self.Me = np.mat(Me)
    
    def getGlobal(self):
        self.getKe()
        self.getNKGe()
        self.getMe()
        
        T = self.T
        '''   COMENTADA PARA ASEGURAR FUNCIONAMIENTO DE BUCKLING
        
        
        # Aplicación de la condensación estática y dinámica de DOF
        R = self.releases
        nR = [not x for x in self.releases]
        
        KFF = self.Ke[np.ix_(nR,nR)]
        KLL = self.Ke[np.ix_(R,R)]
        KFL = self.Ke[np.ix_(nR,R)]
        KLF = self.Ke[np.ix_(R,nR)]
        
        NKGFF = self.NKGe[np.ix_(nR,nR)]
        NKGLL = self.NKGe[np.ix_(R,R)]
        NKGFL = self.NKGe[np.ix_(nR,R)]
        NKGLF = self.NKGe[np.ix_(R,nR)]
        

        
        if(KLL.size != 0):
            Kcon = np.mat(np.zeros((6,6)))
            Kcon[np.ix_(nR,nR)] = KFF - KFL @ np.linalg.solve(KLL,KLF)     
            
            Mcon = np.mat(np.zeros((6,6)))
            """Falta la condensación dinámica aquí"""
            
        elif(NKGLL.size != 0):
  
            NKGcon = np.mat(np.zeros((6,6)))
            NKGcon[np.ix_(nR,nR)] = NKGFF - NKGFL @ np.linalg.solve(NKGLL,NKGLF)  
        
        else:
            Kcon = self.Ke
            NKGcon = self.NKGe
            Mcon = self.Me
        '''
        
        # Cambio a coordenadas globales
        KG = T.T @ self.Ke @ T
        NKGG = T.T @ self.NKGe @ T
        MG = T.T @ self.Me @ T # Falta sustituir la matriz condensada
        
        self.Kg, self.NKGg, self.Mg = np.mat(KG), np.mat(NKGG), np.mat(MG)
        
    
        
    def addLoad(self,L):
        if(type(L).__name__ == "Load" and L.type != "NOD"):
            self.loads.append(L)
            L.addBar(self)
            self.getFemp()
        elif(type(L).__name__ == "list"):
            for Lo in L:
                if(Lo.type != "NOD"):
                    self.loads.append(Lo)
                    Lo.addBar(self)
                    self.getFemp()
                else:
                    raise TypeError("Bar objects do not accept nodal loads")
        else:
            raise TypeError("Bar objects only accept Load objects as loads")
        
    def delLoad(self,ID):
        """Permite eliminar una carga de la barra a través de su ID"""
        if(type(ID).__name__ == 'str' and len(ID) > 0):
            self.loads.rem(ID)
            self.getFemp()
        
    def release(self,dof=[],f=[]):
        """Libera los grados de libertad indicados en dof. Puede contener fuerzas 
        puntuales aplicadas sobre los GDL liberados."""
        
        cont = 1
        self.releases = [False]*6
        self.relF = [0]*6
        
        if(type(dof).__name__ != 'list'):
            dof = [dof]
        if(type(f).__name__ != 'list'):
            f = [f]
            
        if(len(dof) != 0):
            for x in range(len(dof)):
                if(type(dof[x]).__name__ != 'int'):
                    raise TypeError("dof argument must be an integer or list of integers")
                    cont = 0
                elif(dof[x] > 6 or dof[x] < 1):
                    raise ValueError("dof values must be comprised between 1 and 6")
                    cont = 0
                else:
                    self.releases[dof[x]-1] = True
                    if(len(f) >= x+1):
                        try:
                            self.relF[dof[x]] = float(f[x])
                        except ValueError as err:
                            err.args = ('Expected numeric value in forces applied to condensed DOFs; ' + err.args[0],)
                            raise
        
        if(cont == 1):
            self.getGlobal()
            self.getFemp()
        
    def getFemp(self):
        if(len(self.loads) != 0): # Solo ocurren cosas si hay alguna carga metida
            L = self.L
            self.Femp = np.array(np.zeros((6,1)))
            for Lo in self.loads:
                if(not Lo.active):
                    continue
                
                if(Lo.type == "PUN"):
                    a = Lo.a*L
                    b = (1-Lo.a)*L
                    Fx = Lo.fx
                    Fy = Lo.fy
                    Mz = Lo.mz
                    
                    self.Femp += np.array([[-Fx*b/L],
                                           [-Fy*(b**2)*(L+2*a)/(L**3) + 6*Mz*a*b/L**3],
                                           [-Fy*a*(b**2)/L**2 + Mz*a*(1-3*a/L)/L],
                                           [-Fx*a/L],
                                           [-Fy*(a**2)*(L+2*b)/L**3 - 6*Mz*a*b/L**3],
                                           [Fy*(a**2)*b/L**2 + Mz*b*(1-3*b/L)/L]])
    
                elif(Lo.type == "DIS"):
                    a = Lo.a*L
                    b = Lo.b*L
                    qx = Lo.fx
                    mz = Lo.mz
                    
                    Fx1 = -qx*((L-a)**2 - b**2)/(2*L)
                    Fx2 = -qx*((L-b)**2 - a**2)/(2*L)
                    
                    Fz1 = mz*(L**3 - 3*L*a**2 - 3*L*b**2 + 2*a**3 + 2*b**3)/L**3
                    Mz1 = mz*(a*L**2 - 2*L*a**2 - L*b**2 + a**3 + b**3)/L**3
                    Mz2 = Fz1*L - Mz1 - mz*(L - a - b)
                    
                    c = L - a - b
                    a = a + c/2
                    b = b + c/2
                    
                    qy = Lo.fy
                    
                    MA = -qy*c**3*(L - 3*b + 12*a*b**2/c**2)/(12*L**2)
                    MB = qy*c**3*(L - 3*a + 12*b*a**2/c**2)/(12*L**2)
                    
                    self.Femp += np.array([[Fx1],
                                           [-qy*b*c/L - (MA + MB)/L + Fz1],
                                           [MA + Mz1],
                                           [Fx2],
                                           [-qy*a*c/L - (MA + MB)/L - Fz1],
                                           [MB + Mz2]])
    
                elif(Lo.type == "DEF"):
                    ux1 = Lo.ux1
                    uy1 = Lo.uy1
                    th1 = Lo.th1
                    ux2 = Lo.ux2
                    uy2 = Lo.uy2
                    th2 = Lo.th2
                    self.Femp += self.Ke @ np.array([[-ux1],
                                                     [-uy1],
                                                     [-th1],
                                                     [-ux2],
                                                     [-uy2],
                                                     [-th2]])
    
                elif(Lo.type == "TER_T"):
                    T1 = Lo.T1
                    T2 = Lo.T2
                    al = self.mat.al
                    L = self.L
                    h = self.sec.h
                    
                    self.Femp += self.Ke @ np.array([[0],
                                                     [0],
                                                     [0],
                                                     [-al*L*(T1+T2)/2],
                                                     [al*(L**2)*(T1-T2)/(2*h)],
                                                     [al*L*(T1-T2)/h]])
    
                elif(Lo.type == "TER_L"):
                    T1 = Lo.T1
                    T2 = Lo.T2
                    al = self.mat.al
                    L = self.L
                    
                    self.Femp += self.Ke @ np.array([[0],
                                                     [0],
                                                     [0],
                                                     [-al*L*(T1+T2)/2],
                                                     [0],
                                                     [0]])
            # Condensación estática
            R = self.releases
            nR = [not x for x in self.releases]
            
            KLL = self.Ke[np.ix_(R,R)]
            KFL = self.Ke[np.ix_(nR,R)]
            
            FF = self.Femp[np.ix_(nR)]
            
            relF = np.array(self.relF).reshape(6,1)
            FL = relF[np.ix_(R)] + self.Femp[np.ix_(R)]        
            
            if(KLL.size != 0):
                self.FempR = np.array(np.zeros((6,1)))
                self.FempR[np.ix_(nR)] = FF - KFL @ np.linalg.solve(KLL, FL)
                self.FempR[np.abs(self.FempR) < TOL] = 0
            else:
                self.FempR = self.Femp

class BarSol:
    def __init__(self,bar,n1,n2,ug = np.zeros((6,1))):
        
        self.bar = bar
        self.n1 = n1
        self.n2 = n2
        self.ug = ug
        self.ul = np.zeros((6,1))
        self.IF = np.zeros((6,1))
        
        self.getUL()
        self.getIF()
        
    def __str__(self):
        val = 'Solution bar associated to bar ' + str(self.bar.ID) + '\n'
        val = val + 'Local displacements:\n'
        val = val + '    u1  = {: .3e} \n'.format(self.ul[0,0])
        val = val + '    v1  = {: .3e} \n'.format(self.ul[1,0])
        val = val + '    th1 = {: .3e} \n'.format(self.ul[2,0])
        val = val + '    u2  = {: .3e} \n'.format(self.ul[3,0])
        val = val + '    v2  = {: .3e} \n'.format(self.ul[4,0])
        val = val + '    th2 = {: .3e} \n\n'.format(self.ul[5,0])
        val = val + 'Internal forces:\n'
        val = val + '    N1 = {: .3e} \n'.format(self.IF[0,0])
        val = val + '    V1 = {: .3e} \n'.format(self.IF[1,0])
        val = val + '    M1 = {: .3e} \n'.format(self.IF[2,0])
        val = val + '    N2 = {: .3e} \n'.format(self.IF[3,0])
        val = val + '    V2 = {: .3e} \n'.format(self.IF[4,0])
        val = val + '    M2 = {: .3e}'.format(self.IF[5,0])
        
        return val
    
    def __repr__(self):
        return 'Solution bar associated to bar ' + str(self.bar.ID)
    
    def getUL(self):
        T = self.bar.T
        
        self.ul = T @ self.ug
        
        # Cálculo de los GDL condensados
        R = self.bar.releases
        nR = [not x for x in self.bar.releases]
        
        KLL = self.bar.Ke[np.ix_(R,R)]
        
        if(KLL.size != 0):
            KLF = self.bar.Ke[np.ix_(R,nR)]
            
            relF = np.array(self.bar.relF).reshape(6,1)
            FL = relF[np.ix_(R)] - self.bar.Femp[np.ix_(R)] 
            
            uF = self.ul[np.ix_(nR)]
            
            self.ul[np.ix_(R)] = np.linalg.solve(KLL, FL - KLF@uF)
            self.ul[np.abs(self.ul) < TOL] = 0
            self.ug = T.T @ self.ul
            self.ug[np.abs(self.ug) < TOL] = 0
            
    def getIF(self):
        self.IF = self.bar.Ke @ self.ul + self.bar.Femp
        self.IF[np.abs(self.IF) < TOL] = 0
        
        # Inclusión de las fuerzas puntuales en los esfuerzos
        R = self.bar.releases
        
        KLL = self.bar.Ke[np.ix_(R,R)]
        
        if(KLL.size != 0):
            self.IF += np.array(self.bar.relF).reshape((6,1))
            
class BarList(list):
    def __init__(self,bars=[]):
        if(type(bars).__name__ in ['Bar','list','BarList']):
            list.__init__(self,[])
            if type(bars).__name__ == 'Bar':
                self.append(bars)
            elif type(bars).__name__ in ['list','BarList']:
                for B in bars:
                    if B.dupli or (not B.dupli and not B in self):
                        self.append(B)
                    else:
                        raise Exception('Bar {} already exists and cannot be duplicated.'.format(B.ID))
            
    def __repr__(self):
        return str([a for a in self])
    
    def __contains__(self, key):
        if(type(key).__name__ == 'Bar'):
            is_in = 0
            for b in self:
                if(key.p1 == b.p1 and key.p2 == b.p2):
                    is_in += 1
                
            if(is_in > 0):
                return True
            else:
                return False
        else:
            print(key,'is not a valid Bar object')
            return False
        
    def p1(self):
        return [I.p1 for I in self]
    
    def p2(self):
        return [I.p2 for I in self]
    
    def L(self):
        return [I.L for I in self]
    
    def th(self):
        return [I.th for I in self]
    
    def mat(self):
        return [I.mat for I in self]
    
    def sec(self):
        return [I.sec for I in self]
    
    def has(self, node):
        if(type(node).__name__ == 'Node'):
            if(len(self) > 0):
                for L in self:
                    if(L.x == node.x and L.y == node.y and L.th == node.th and L.ID == node.ID):
                        return True
            else:
                return False
        else:
            print('Invalid type')
        
        return False