# -*- coding: utf-8 -*-
"""
Created on Tue Dec 12 21:19:13 2017

@author: Alvaro
"""

from matplotlib.collections import PatchCollection
from matplotlib import patches as ptch
from matplotlib.path import Path as pth
from matplotlib.transforms import Affine2D

import numpy as np
from math import sin, cos

class _BC(PatchCollection):
    def __init__(self,node, *args, **kwargs):
        PatchCollection.__init__(self, *args, **kwargs, 
                            edgecolors = '#1e488f',
                            facecolors = '#95d0fc',
                            lw = 1.5,
                            zorder = 3)
        
        self.node = node
        self.ax = None
    
    _th = 0 # Angle rotation of the drawing
    _sc = 1 # Scale factor
        
    @property
    def th(self):
        """Angle of the graphical object."""
        return self._th
    
    @th.setter
    def th(self, th):
        """Set the angle and rotates the graphical object 'on the go'"""
        self.set_transform(Affine2D.identity().rotate_around(self.node.x, self.node.y, th)
                                              .scale(self._sc)
                                              .translate(self.node.x*(1-self._sc), self.node.y*(1-self._sc)) 
                                              + self.ax.transData)
        self._th = th
        
    @property
    def sc(self):
        """Scale factor of the graphical object."""
        return self._sc
    
    @sc.setter
    def sc(self, sc):
        """Set the scale factor and scales the graphical object 'on the go'"""
        self.set_transform(Affine2D.identity().rotate_around(self.node.x, self.node.y, self._th)
                                              .scale(sc)
                                              .translate(self.node.x*(1-sc), self.node.y*(1-sc))
                                              + self.ax.transData)
        self._sc = sc

class Clamped(_BC):
    """Graphic object associated with a clamped BC."""
    def __init__(self, node, **kwargs):
        _BC.__init__(self, node, [])
        
        base = ptch.Polygon(((node.x-0.9*self.sc, node.y),(node.x+0.9*self.sc, node.y)), closed = False)
        y0 = node.y - 0.5*0.5*self.sc
        y1 = node.y
        x1 = [node.x-0.9*self.sc, node.x-0.9*self.sc/2, node.x, node.x+0.9*self.sc/2, node.x+0.9*self.sc]
        x0 = [x-0.866*0.5*0.5*self.sc for x in x1]
        p = [ptch.Polygon(((x0[i], y0),(x1[i],y1)), closed = False) for i in range(5)]
        
        self.set_paths([base]+p)
        
        
class Fixed(_BC):
    """Graphic object associated with a fixed support BC."""
    def __init__(self, node, **kwargs):
        
        _BC.__init__(self, node, [])
        
        tri = ptch.RegularPolygon((node.x, node.y - self.sc), 3, self.sc)
        y0 = node.y - self.sc - 0.5*self.sc - 0.5*0.5*self.sc
        y1 = node.y - self.sc - 0.5*self.sc
        x1 = [node.x-self.sc*0.866, node.x-self.sc*0.433, node.x, node.x+self.sc*0.433, node.x+self.sc*0.866]
        x0 = [x-0.866*0.5*0.5*self.sc for x in x1]
        p = [ptch.Polygon(((x0[i], y0),(x1[i],y1)), closed = False) for i in range(5)]
        
        self.set_paths([tri]+p)
        
        
class Moving(_BC):
    """Graphic object associated with a moving support BC."""
    def __init__(self, node, **kwargs):
        _BC.__init__(self, node, [])
        #Triángulo
        tri = ptch.RegularPolygon((node.x, node.y - 0.8*self.sc), 3, 0.8*self.sc)
        
        #Ruedas
        yC = node.y - 1.35*self.sc
        xC = [node.x-0.75*0.866*0.8*self.sc, node.x-0.25*0.866*0.8*self.sc,
              node.x+0.25*0.866*0.8*self.sc, node.x+0.75*0.866*0.8*self.sc]
        r = [ptch.Circle((xC[i], yC), 0.15*self.sc) for i in range(len(xC))]
        
        #Suelo
        base = ptch.Polygon(((node.x-0.866*0.8*self.sc, node.y - 1.5*self.sc),
                             (node.x+0.866*0.8*self.sc, node.y - 1.5*self.sc)), 
                             closed = False)
        y0 = node.y - 0.5*0.5*self.sc - 1.5*self.sc
        y1 = node.y - 1.5*self.sc
        x1 = [node.x-0.8*0.866*self.sc, 
              node.x-0.8*0.866*self.sc/2, 
              node.x, 
              node.x+0.8*0.866*self.sc/2, 
              node.x+0.8*0.866*self.sc]
        x0 = [x-0.866*0.5*0.5*self.sc for x in x1]
        p = [ptch.Polygon(((x0[i], y0),(x1[i],y1)), closed = False) for i in range(5)]
        
        
        self.set_paths([tri, base]+r+p)

class MovClamped(_BC):
    """Graphic object associated with a moving clamped BC."""
    def __init__(self, node, **kwargs):
        _BC.__init__(self, node, [])
        # Linea
        linea = ptch.Polygon(((node.x-0.8*self.sc, node.y),
                             (node.x+0.8*self.sc, node.y)), 
                             closed = False)
        #Ruedas
        yC = node.y - 0.15*self.sc
        xC = [node.x-0.75*0.8*self.sc, node.x-0.25*0.8*self.sc,
              node.x+0.25*0.8*self.sc, node.x+0.75*0.8*self.sc]
        r = [ptch.Circle((xC[i], yC), 0.15*self.sc) for i in range(len(xC))]
        
        #Suelo
        base = ptch.Polygon(((node.x-0.8*self.sc, node.y - 0.3*self.sc),
                             (node.x+0.8*self.sc, node.y - 0.3*self.sc)), 
                             closed = False)
        y0 = node.y - 0.5*0.5*self.sc - 0.3*self.sc
        y1 = node.y - 0.3*self.sc
        x1 = [node.x-0.8*self.sc, node.x-0.8*self.sc/2, node.x, node.x+0.8*self.sc/2, node.x+0.8*self.sc]
        x0 = [x-0.866*0.5*0.5*self.sc for x in x1]
        p = [ptch.Polygon(((x0[i], y0),(x1[i],y1)), closed = False) for i in range(5)]
        
        self.set_paths([linea, base]+r+p)
        
class NoRotation(_BC):
    """Graphic object associated with an only-rotation-restricted BC."""
    def __init__(self, node, **kwargs):
        _BC.__init__(self,  node, [])
        
        linea1 = ptch.Polygon(((node.x-0.6*self.sc, node.y),
                             (node.x+0.6*self.sc, node.y)), 
                             closed = False)
        linea2 = ptch.Polygon(((node.x-0.3*self.sc, node.y-0.15*self.sc),
                             (node.x, node.y)), 
                             closed = False)
        linea3 = ptch.Polygon(((node.x, node.y-0.3*self.sc),
                             (node.x, node.y)), 
                             closed = False)
        linea4 = ptch.Polygon(((node.x+0.3*self.sc, node.y-0.15*self.sc),
                             (node.x, node.y)), 
                             closed = False)
        self.set_paths([linea1, linea2, linea3, linea4])