# -*- coding: utf-8 -*-
"""
Created on Wed Aug 30 19:42:28 2017
@author:  Alvaro Magdaleno Gonzalez

Modified on Jun 2021 by Jose Manuel Martin Cabeza
"""
import numpy as np
import matplotlib.pyplot as plt
import warnings
from math import sqrt, sin, cos, pi, atan

from StruPy2D.config import TOL

class Node:
    def __init__(self,x=0,y=0,th=0):
        self.ID = ''
        
        try: # Comprobación de los argumentos de entrada
            self.x = float(x)
            self.y = float(y)
            self.th = float(th)
        except ValueError as err:
            err.args = ('Expected numeric values when a Node is created; ' + err.args[0],)
            raise
            
        self.bars = []
        self.bc = [True,True,True] # Boundary conditions
        self.k = [0,0,0] # Stiffnesses
        self.u = [0,0,0] # Displacements
        self.loads = [] # Nodal load list
        self.dupli = False
        
    def __str__(self):
        print("(x, y) = (",self.x,", ",self.y,")",sep='')
        
        if(self.th != 0):
            print("th = ",self.th*180/pi," degrees",sep='')
        if(len(self.bars) > 0):
            print("It belongs to bars:",[B.ID for B in self.bars])
            
        print("Boundary conditions:")
        sim1 = ['ux','uy','th']
        sim2 = ['kx','ky','kth']
        sim3 = ['dx','dy','dth']
        
        for i in range(3):
            if(self.bc[i] == 1):
                print("     ",sim1[i],": free,        ",sim2[i]," = ", self.k[i],sep='')
            else:
                print("     ",sim1[i],": constrained, ",sim3[i]," = ", self.u[i],sep='')
        
        if(len(self.loads) != 0):
            i = 1
            print('It is loaded with the following Nodal forces:')
            for L in self.loads:
                print("Load number ",i,":",sep='')
                print("   Fx = ",L.fx,sep='')
                print("   Fy = ",L.fy,sep='')
                print("   Mz = ",L.mz,sep='')
                i += 1
                
        return ''
    
    def __repr__(self):
        if(len(self.ID) > 0):
            return str(self.ID)
        else:
            return '({},{})'.format(self.x,self.y)
    
    def __eq__(self, key):
        if(type(key).__name__ != 'Node'):
            raise TypeError('Expected Node object for comparison.')
        
        if(key.x == self.x and key.y == self.y):
            return True
        else:
            return False
                
    def __ne__(self, key):
        if(type(key).__name__ == 'Node'):
            raise TypeError('Expected Node object for comparison.')
            
        if(key.x != self.x or key.y != self.y):
            return True
        else:
            return False
    
    def relocate(self, x = None, y = None, th = None):
        try:
            if(not x is None):
                self.x = float(x)
            if(not y is None):
                self.y = float(y)
            if(not th is None):
                self.th = float(th)
        except TypeError as err:
            err.args = ('Expected numeric values when a Node is relocated; ' + err.args[0],)
            raise
        
        if len(self.bars) > 0:
            for b in self.bars:
                b.updateLth()
                
    def move(self,dx=0,dy=0):
        try:
            self.x += float(dx)
            self.y += float(dy)
        except TypeError as err:
            err.args = ('Expected numeric values when a Node is moved; ' + err.args[0],)
            raise
        
        if len(self.bars) > 0:
            for b in self.bars:
                b.updateLth()
                
    def rotate(self, dth=0):
        try:
            self.th += dth
        except TypeError as err:
            err.args = ('Expected numeric values when a Node is moved; ' + err.args[0],)
            raise
        
    def dist(self,p):
        if(type(p).__name__ != 'Node'):
            raise TypeError("When calculating a distance between two Nodes, input argument must be a Node instance.")
        
        return sqrt((self.x - p.x)**2 + (self.y - p.y)**2)
    
    def setBar(self,b):            
        self.bars.append(b)
    
    def setBC(self,type={},**args):
        """ Set boundary conditions to the node: ux, uy, th; kx, ky, kth; dx, dy, dth"""
        """ It doesn't add a BC, but restart them all from their default values"""
        type = type.copy()
        
        if len(args) > 0:
            for entry in args:
                type[entry] = args[entry]
        
        ke = type.keys()
        
        dis_vars = ['ux','uy','th']
        bc_vars = ['kx','ky','kth','dx','dy','dth']
        
        for K in ke:
            if(not K in dis_vars and not K in bc_vars):
                warnings.warn(str(K) + ' is not a valid Boundary Condition argument',SyntaxWarning)
            
        for dv in dis_vars:
            if dv not in type:
                type[dv] = 1
        for bv in bc_vars:
            if bv not in type:
                type[bv] = 0.0
        
        try:
            self.bc = [type['ux']==1,type['uy']==1,type['th']==1]
            self.k = [float(type['kx']),float(type['ky']),float(type['kth'])]
            self.u = [float(type['dx']),float(type['dy']),float(type['dth'])]
        except TypeError as Terr:
            Terr.args = ('Expected numeric (1/0) or boolean variable when a BC is set; ' + Terr.args[0],)
            raise
            
        except ValueError as Verr:
            Verr.args = ('Expected numeric (1/0) or boolean variable when a BC is set; ' + Verr.args[0],)
            raise
        
    def modBC(self,**args):
        """It updates the boundary conditions without modifying the others"""
        if len(args) > 0:
            for entry in args:
                if(entry not in ['ux','uy','th','kx','ky','kth','dx','dy','dth']):
                    warnings.warn(str(entry) + ' is not a valid Boundary Condition argument',SyntaxWarning)
                    
                if entry == 'ux':
                    self.bc[0] = args[entry] == 1
                elif entry == 'uy':
                    self.bc[1] = args[entry] == 1
                elif entry == 'th':
                    self.bc[2] = args[entry] == 1
                elif entry == 'kx':
                    try:
                        self.k[0] = float(args[entry])
                    except TypeError as err:
                        err.args = ('Expected numeric variable when setting kx; ' + err.args[0],)
                        raise
                elif entry == 'ky':
                    try:
                        self.k[1] = float(args[entry])
                    except TypeError as err:
                        err.args = ('Expected numeric variable when setting ky; ' + err.args[0],)
                        raise
                elif entry == 'kth':
                    try:
                        self.k[2] = float(args[entry])
                    except TypeError as err:
                        err.args = ('Expected numeric variable when setting kth; ' + err.args[0],)
                        raise
                elif entry == 'dx':
                    try:
                        self.u[0] = float(args[entry])
                    except TypeError as err:
                        err.args = ('Expected numeric variable when setting dx; ' + err.args[0],)
                        raise
                elif entry == 'dy':
                    try:
                        self.u[1] = float(args[entry])
                    except TypeError as err:
                        err.args = ('Expected numeric variable when setting dy; ' + err.args[0],)
                        raise
                elif entry == 'dth':
                    try:
                        self.u[2] = float(args[entry])
                    except TypeError as err:
                        err.args = ('Expected numeric variable when setting dth; ' + err.args[0],)
                        raise
            
    def addLoad(self,L):
        if(type(L).__name__ == "Load" and L.type == "NOD"):
            self.loads.append(L)
            L.addNode(self)
        elif(type(L).__name__ == "list"):
            for LL in L:
                if(LL.type == 'NOD'):
                    self.loads.append(LL)
                    LL.addNode(self)
                else:
                    raise TypeError("The Load type is not valid for a Node object.")
        elif(type(L).__name__ == "Load" and L.type != "NOD"):
            raise TypeError("The Load type is not valid for a Node object.")
        else:
            raise TypeError("The object is not a valid Load or List object.")
        
    
    def copy(self,dx=0,dy=0,dth=0):
        """Retorna un Node exactamente igual a este, desplazado una cantidad dx
        dy"""
        a = Node(self.x,self.y,self.th)
        a.bars = self.bars
        a.bc = self.bc
        a.k = self.k
        a.u = self.u
        a.loads = self.loads
        try:
            a.move(dx,dy)
            a.rotate(dth)
        except TypeError as err:
            err.args = ('Expected numeric values when a Node is copied; ' + err.args[0],)
            raise
            
        return a
    
    def offset(self,dx=0,dy=0,dth=0):
        """Retorna un Node vacío, pero de coordenadas iguales a este más un
        offset dado por dx,dy"""
        a = Node(self.x,self.y)
        try:
            a.move(dx,dy)
            a.rotate(dth)
        except TypeError as err:
            err.args = ('Expected numeric values when a Node is offset; ' + err.args[0],)
            raise
            
        return a
    
    """
    def show(self,ax=[]):
        if(len(ax) == 0):
            fig, ax = plt.subplots()
            ax.plot(self.x,self.y,'ob')
            
            if(len(self.loads) > 0):
                drawLoad(ax,self)
            
            if(np.prod(self.bc) == 0):
                drawBC(ax,self)
              
            ax.axis([self.x-1.05,self.x+1.05,self.y-1.05,self.y+1.05])
            ax.margins(0.05)
            
        elif(type(ax).__name__ in ['AxesSubplot','Axes']):
            ax.plot(self.x, self.y, 'ob')
        else:
            print("Axes object not valid")
        """
    
        
class NodeSol: # Sin excepciones porque no está previsto su uso público
    def __init__(self,node,ug=np.zeros((3,1)),Fg=np.zeros((3,1)),ubug=np.zeros((3,1))):
        self.node = node
        self.ug = ug
        self.ul = np.zeros((3,1))
        self.Fg = Fg
        self.Fl = np.zeros((3,1))
        self.ubug = ubug #Almacena los desplazamientos en coordenadas globales por pandeo
        
        self.getLocal()
        
    def __str__(self):
        val = 'Solution node associated to node ' + str(self.node.ID) + '\n'
        val = val + 'Local displacements:\n'
        val = val + '    u  = {: .3e} \n'.format(self.ul[0,0])
        val = val + '    v  = {: .3e} \n'.format(self.ul[1,0])
        val = val + '    th = {: .3e} \n\n'.format(self.ul[2,0])
        val = val + 'External forces in local coordinates:\n'
        val = val + '    Fx = {: .3e} \n'.format(self.Fl[0,0])
        val = val + '    Fy = {: .3e} \n'.format(self.Fl[1,0])
        val = val + '    Mz = {: .3e}'.format(self.Fl[2,0])

        return val
    
    def __repr__(self):
        return 'Solution node associated to node ' + str(self.node.ID)
    
    def getLocal(self):
        th = self.node.th
        L = np.mat([[cos(th), sin(th), 0],[-sin(th), cos(th), 0],[0, 0, 1]])
        self.ul = L @ self.ug
        self.Fl = L @ self.Fg
        
class NodeList(list): # Sin excepciones porque no está previsto su uso público
    def __init__(self,nds=[]):
        if(type(nds).__name__ in ['Node','list','NodeList']):
            list.__init__(self,nds)
            
    def __repr__(self):
        return str([a for a in self])
    
    def __contains__(self, key):
        if(type(key).__name__ == 'Node'):
            for N in self:
                if(N.x == key.x and N.y == key.y):
                    return True
            
            return False
        else:
            # Meter una excepción de tipos
            raise TypeError(str(key)+' is not a valid Node object')
            
    def x(self):
        return [I.x for I in self]
    
    def y(self):
        return [I.y for I in self]
    
    def th(self):
        return [I.th for I in self]
    
    def has(self, node):
        if(type(node).__name__ == 'Node'):
            if(len(self) > 0):
                for L in self:
                    if(L.x == node.x and L.y == node.y and L.th == node.th and L.ID == node.ID):
                        return True
            else:
                return False
        else:
            print('Invalid type')
        
        return False