# -*- coding: utf-8 -*-
"""
Created on Wed Aug 30 19:42:43 2017

@author: Alvaro

Modified on Jun 2021 by Jose Manuel Martin Cabeza
"""

import numpy as np
from math import sqrt, sin, cos, pi, atan
from matplotlib.axes._axes import Axes as mpAxes
import matplotlib.pyplot as plt
import matplotlib.patches as patches

from StruPy2D.drawBC import drawBC, drawElasSup, drawDispSup
from StruPy2D.drawLoad import drawLoad
import StruPy2D.Graphics.Axes as spAxes
from StruPy2D.Graphics.graphNode import graphNode as gN
from StruPy2D.Graphics.graphBar import graphBar as gB
from StruPy2D.Graphics.graphLoad import graphLoad as gL

from StruPy2D.config import TOL
from StruPy2D.Load import Load, LoadGroup, LoadList
from StruPy2D.Node import NodeList
from StruPy2D.Bar import BarList
from StruPy2D.Material import MatList, defMat
from StruPy2D.Section import SecList, defSec
from StruPy2D.Graphics.Style import defNodeStyle, defBarStyle, defLoadStyle

class Structure:
    def __init__(self,bars=[],nds=[],mats=[],secs=[],loads=[],damp=[]):
        """Inicialización de la estructura. Opcional: un conjunto de puntos,
        barras, materiales y secciones. Las barras atraerán puntos, pero no al
        revés, perdiéndose la jerarquía. Asimismo, las barras atraen materiales
        y secciones, pero no al revés. También atraerán cargas. Los puntos 
        traen sus cargas. Las cargas no añaden elementos estructurales. """
        self.nds = NodeList([])
        self.bars = BarList([])
        self.mats = MatList([])
        self.secs = SecList([])
        self.loads = LoadList([])
        
        self.damp = damp
        
        self._weight = LoadGroup('_weight') # Grupo de cargas peso propio
        self.LG = [] # Lista de LoadGroup inicializada a 0
        self.g = 9.81 # Aceleración de la gravedad
        
        for B in bars:
            self.addBar(B)
            
        for N in nds:
            self.addNode(N)
        
        for M in mats:
            self.addMaterial(M)
            
        for S in secs:
            self.addSection(S)
            
        for L in loads:
            self.addLoad(L)
        
        self.getMatrices()
        self.getFeq()
        
    def __str__(self):
        return 'Structure with ' + str(len(self.bars)) + ' bars.'
    
    def __repr__(self):
        return 'Main structure'
    
    def __contains__(self, obj):
        """Evalúa si el objeto existe en la estructura."""
        if(type(obj).__name__ == 'Node'):
            return obj in self.nds
        elif(type(obj).__name__ == 'Bar'):
            return obj in self.bars
        elif(type(obj).__name__ == 'Material'):
            pass
        elif(type(obj).__name__ == 'Section'):
            pass
        elif(type(obj).__name__ == 'Load'):
            pass
        else:
            return TypeError("Unknown object")
            
        
    def addNode(self,p):
        """Añade un punto a la lista de puntos. Verificar si pueden existir duplicados"""
        if(type(p).__name__ == "Node"):
            if(p.dupli or (not p.dupli and not p in self.nds)):
                self.nds.append(p)
                for L in p.loads:
                    if(not L in self.loads):
                        self.loads.append(L)
            else:
                print("The point "+repr(p)+" already exists in the structure and cannot be duplicated.")
                
        elif(type(p).__name__ == "list"):
            for P in p:
                if(type(P).__name__ == "Node"):
                    self.nds.append(P)
                    for L in P.loads:
                        if(not L in self.loads):
                            self.loads.append(L)
                else:
                    raise TypeError("List of Nodes expected.")
        else:
            raise TypeError("Node or list object expected.")

        
    def delNode(self,n):
        """Elimina uno o varios nodos de la estructura. Por supuesto, las 
        barras que confluyan en estos nodos desaparecen con ellos, no así las
        cargas, que solo pierden su referencia."""
        if(type(n).__name__ == "Node"):
            if(not n in self.nds):
                raise Exception('The node is not in the Structure')
                
            for L in n.loads:
                L.nodes.remove(n)
                
            self.nds.remove(n)
            
            self.delBar(n.bars,_rec=True)
            # del(n)
        elif(type(n).__name__ == "list"):
            for N in n.copy():
                if(N in self.nds):
                    try:
                        self.delNode(N)
                    except TypeError as err:
                        err.args = ['Node object expected in list.']
                        raise err
                        
        else:
            raise TypeError("Node or list object expected.")
        
        self.getMatrices()
        self.getFeq()
        
        
    def addBar(self,b,_rec=False):
        """Añade una barra a la lista de barra, con todo lo que ello conlleva 
        sobre la adición de materiales, secciones y puntos. Solo se añaden los
        puntos que no existan ya"""      
        if(type(b).__name__ == "Bar"):              
            if((b.dupli and not b in self.bars) or (not b.dupli and not b in self.bars)):
                self.bars.append(b)
                
                if(not b.p1 in self.nds):
                    self.nds.append(b.p1)
                if(not b.p2 in self.nds):
                    self.nds.append(b.p2)
                if(not b.mat in self.mats):
                    self.mats.append(b.mat)
                if(not b.sec in self.secs):
                    self.secs.append(b.sec)
                for L in b.loads:
                    if(not L in self.loads):
                        self.loads.append(L)
                for L in b.p1.loads:
                    if(not L in self.loads):
                        self.loads.append(L)
                for L in b.p2.loads:
                    if(not L in self.loads):
                        self.loads.append(L)
                
                self.addWeight(b)
                
            elif(b.dupli and b in self.bars):
                self.bars.append(b)
                self.addWeight(b)
                b.p1.setBar(b)
                b.p2.setBar(b)
                    
            else:
                print("The bar "+repr(b)+" already exists in the structure and cannot be duplicated.")
                
        elif(type(b).__name__ == "list"):
            for bar in b:
                try:
                    self.addBar(bar,True)
                except TypeError as err:
                    err.args = ['Bar object expected inside List']
                    raise err
        else:
            raise TypeError("Bar or list object expected.")
        
        if(not _rec):
            self.getMatrices()
            self.getFeq()
        
        
    def delBar(self,b,_rec=False):
        """Elimina una barra de la estructura, junto a todas sus dependencias. 
        Los nodos huérfanos no se eliminan aquí; hay que hacer una purga."""
        if(type(b).__name__ == "Bar"):
            if(not b in self.bars):
                raise Exception('The bar is not in the Structure')
                
            b.p1.bars.remove(b)
            b.p2.bars.remove(b)
            for L in b.loads:
                L.bars.remove(b)
            
            # Se quita la carga de peso propio de esa barra
            self.delWeight(b)
            # Quitamos la barra
            self.bars.remove(b)
            
        elif(type(b).__name__ == "list"):
            for B in b.copy():
                try:
                    self.delBar(B,True)
                except TypeError as err:
                    err.args = ['Bar object expected inside List']
                    raise err
        else:
            raise TypeError("Bar or list object expected.")
        
        if(not _rec): # Si no es una llamada recursiva...
            self.getMatrices()
            self.getFeq()
        
        
    def addMaterial(self,mat):
        """Añade un material nuevo a la estructura. Se comprueba que no exista
        ya y, de no ser así, se añade. Se comprueba además que no exista 
        ningún otro material con el mismo nombre."""        
        mat_list = [m.ID for m in self.mats]
        
        if(type(mat).__name__ == 'Material'):
            if(mat.ID in mat_list):
                raise Exception('The material has been already created.')
                
            self.mats.append(mat)
            
        elif(type(mat).__name__ == 'list'):
            for M in mat:
                try:
                    self.addMaterial(M)
                except TypeError as err:
                    err.args = ['Material object expected inside List']
                    raise err
                
        else:
            raise TypeError("Material or list object expected.")
        
    
    def delMaterial(self, mat, _rec=False):
        """Elimina un material de la lista de materiales y asigna a todas las
        barras que lo tuvieran asignado un material por defecto."""
        mat_list = [m.ID for m in self.mats]
        
        if(type(mat).__name__ == 'Material'):
            if(not mat.ID in mat_list):
                raise Exception('The material does not exist in the Structure.')
                
            self.mats.remove(mat)
            for b in self.bars:
                if(b.mat.ID == mat.ID):
                    b.mat = defMat
                    b.getGlobal()
                    b.getFemp()
                    
        elif(type(mat).__name__ == 'list'):
            for M in mat.copy():
                try:
                    self.delMaterial(M,True)
                except TypeError as err:
                    err.args = ['Material object expected inside List']
                    raise err
                
        else:
            raise TypeError("Material or list object expected.")
        
        if(not _rec):
            self.getMatrices()
            self.getFeq()
        
    def addSection(self, sec):
        """Añade una nueva sección a la estructura. Se comprueba que no exista 
        ya y, de noser así, se añade. """
        sec_list = [s.ID for s in self.secs]
        
        if(type(sec).__name__ == 'Section'):
            if(sec.ID in sec_list):
                raise Exception('The section has been already created.')
                
            self.secs.append(sec)
            
        elif(type(sec).__name__ == 'list'):
            for S in sec:
                try:
                    self.addSection(S)
                except TypeError as err:
                    err.args = ['Section object expected inside List']
                    raise err
                
        else:
            raise TypeError("Section or list object expected.")
        
    def delSection(self,sec,_rec=False):
        """Elimina un material de la lista de materiales y asigna a todas las
        barras que lo tuvieran asignado un material por defecto."""
        sec_list = [s.ID for s in self.secs]
        
        if(type(sec).__name__ == 'Section'):
            if(not sec.ID in sec_list):
                raise Exception('The section does not exist in the Structure.')
                
            self.secs.remove(sec)
            for b in self.bars:
                if(b.sec.ID == sec.ID):
                    b.sec = defSec
                    b.getGlobal()
                    b.getFemp()
                    
        elif(type(sec).__name__ == 'list'):
            for S in sec.copy():
                try:
                    self.delSection(S,True)
                except TypeError as err:
                    err.args = ['Section object expected inside List']
                    raise err
                
        else:
            raise TypeError("Section or list object expected.")
        
        if(not _rec):
            self.getMatrices()
            self.getFeq()
        
    def addLoad(self,load):
        """Añade una fuerza en la lista de cargas de la estructura."""
        load_list = [l.ID for l in self.loads]
        
        if(type(load).__name__ == 'Load'):
            if(load.ID in load_list):
                raise Exception('The load has been already created.')
                
            self.loads.append(load)
            
        elif(type(load).__name__ == 'list'):
            for L in load:
                try:
                    self.addLoad(L)
                except TypeError as err:
                    err.args = ['Load object expected inside List']
                    raise err
                
        else:
            raise TypeError("Load or list object expected.")
            
    def delLoad(self,load, _rec=False):
        """Elimina una carga de la lista de cargas de la estructura."""
        load_list = [l.ID for l in self.loads]
        
        if(type(load).__name__ == 'Load'):
            if(not load.ID in load_list):
                raise Exception('The load does not exist in the Structure.')
                
            self.loads.remove(load)
            if(load.type == 'NOD'):
                for N in self.nds:
                    for L in N.loads:
                        if(L.ID == load.ID):
                            N.loads.remove(L)
                            
            else:    
                for B in self.bars:
                    for L in B.loads:
                        if(L.ID == load.ID):
                            B.loads.remove(L)
                            B.getFemp()
                            
        elif(type(load).__name__ == 'list'):
            for L in load.copy():
                try:
                    self.delLoad(L,True)
                except TypeError as err:
                    err.args = ['SLoad object expected inside List']
                    raise err
                    
        else:
            raise TypeError("Load or list object expected.")
            
        if(not _rec):
            self.getMatrices()
            self.getFeq()
        
    def addLoadGroup(self, LG):
        """Se añade un grupo de cargas a la lista de grupos de cargas de la
        estructura."""
        LG_list = [lg.ID for lg in self.LG]
        
        if(type(LG).__name__ == 'LoadGroup'):
            if(LG.ID in LG_list):
                raise Exception('The Load Group has been already created.')
                
            self.LG.append(LG)
            
        elif(type(LG).__name__ == 'list'):
            for lg in LG:
                try:
                    self.addLoadGroup(lg)
                except TypeError as err:
                    err.args = ['LoadGroup object expected inside List']
                    raise err
                
        else:
            raise TypeError("LoadGroup or list object expected.")
            
    def delLoadGroup(self, LG):
        """Se elimina un grupo de cargas de la lista de grupos de cargas,
        previa activación de todas las cargas que contiene."""
        LG_list = [lg.ID for lg in self.LG]
        
        if(type(LG).__name__ == 'LoadGroup'):
            if(not LG.ID in LG_list):
                raise Exception('The Load Group does not exist.')
                
            if(LG.ID == '_weight'):
                raise Exception('Self weight LoadGroup cannot be deleted')
                
            self.LG.turn('on')
            self.LG.remove(LG)
            
        elif(type(LG).__name__ == 'list'):
            for lg in LG.copy():
                try:
                    self.delLoadGroup(lg)
                except TypeError as err:
                    err.args = ['LoadGroup object expected inside List']
                    raise err
                
        else:
            raise TypeError("LoadGroup or list object expected.")
    
    def addWeight(self, bar):
        """Incluye el peso propio de la barra 'bar' en el LoadGroup _weight"""
        
        if(type(bar).__name__ == 'Bar'):
            id_list = [W.ID for W in self._weight.loads]
            ID = '_w_{}'.format(bar.ID)
            
            if(not ID in id_list):
                Fx = -bar.mat.rho*self.g*bar.sec.A*sin(bar.th)
                Fy = -bar.mat.rho*self.g*bar.sec.A*cos(bar.th)
                L = Load(ID,'DIS',fx = Fx, fy = Fy)
                bar.addLoad(L)
                self._weight.addLoad(L)
        else:
            raise TypeError('Expected Bar object')
        
    def delWeight(self, bar):
        """Quita el peso propio de la barra 'bar' del LoadGroup _weight"""
        
        if(type(bar).__name__ == 'Bar'):
            id_list = [W.ID for W in self._weight.loads]
            ID = '_w_{}'.format(bar.ID)
            
            if(ID in id_list):
                bar.delLoad(ID)
                self._weight.delLoad(ID)
        else:
            raise TypeError('Expected Bar object')
            
    def setWeight(self, mode):
        """Activación o desactivación del peso propio"""
        
        if(type(mode).__name__ == 'str'):
            if(mode in ['on','On','ON'] and not self._weight.active):
                self._weight.turn('on')
                
            elif(mode in ['off','Off','OFF'] and self._weight.active):
                self._weight.turn('off')
                
    def setGravity(self, g):
        """Cambia el valor de la aceleración de la gravedad para el peso."""
        try:
            g = float(g)
        except ValueError as err:
            err.args = ['Gravity acceleration must be a number']
            raise err
            
        old_g= self.g
        self.g = g
        
        for W in self._weight.loads:
            W.fx = W.fx*g/old_g
            W.fy = W.fy*g/old_g
            
        for b in self.bars:
            b.getFemp()
        
    def cleanNodes(self):
        """Elimina los nodos que no pertenezcan a ninguna barra."""
        for N in self.nds:
            if(len(N.bars) == 0):
                self.nds.remove(N)
                
    def reorderNodes(self):
        """Algoritmo de reordenación de nodos, para reducir el ancho de banda
        de las matrices de la estructura. Primero se colocan los nodos que no
        tienen barras asociadas al final de la lista. Luego, de los restantes, 
        se va siguiendo la conectividad."""
        
        self.cleanNodes()
        
    def assembleKNKGM(self):
        """Ensamblaje de las matrices de masa y rigidez de la estructura."""
        self.reorderNodes()
        
        self.Kg = np.mat(np.zeros((3*len(self.nds),3*len(self.nds))))
        self.NKGg = np.mat(np.zeros((3*len(self.nds),3*len(self.nds))))
        self.Mg = np.mat(np.zeros((3*len(self.nds),3*len(self.nds))))

        
        for B in self.bars:
            B.getGlobal()
            p1 = self.nds.index(B.p1)
            p2 = self.nds.index(B.p2)
            
            ind = np.array([np.arange(3*p1,3*p1+3,1),
                             np.arange(3*p2,3*p2+3,1)]).reshape(6,1)
            
            self.Kg[ind,ind.T] += B.Kg
            self.NKGg[ind,ind.T] += B.NKGg 
            self.Mg[ind,ind.T] += B.Mg
            
               
    def getNodalMatrices(self): 
        """Función que proporciona las matrices nodales a partir de las
        globales de la estructura"""
        
        for N in self.nds:
            if N.th != 0: # Solo intenta cambio de base si no coinciden los ejes
                th = N.th
                L = np.mat([[cos(th),sin(th),0],[-sin(th),cos(th),0],[0,0,1]])
                
                i = self.nds.index(N)
                ind = np.array([3*i,3*i+1,3*i+2]).reshape(3,1)
                
                tam = self.Kg.shape[0]
                ind2 = np.array(np.arange(tam).reshape((tam,1)))
                
                self.Kg[ind,ind2.T] = L @ self.Kg[ind,ind2.T]
                self.Kg[ind2,ind.T] = self.Kg[ind2,ind.T] @ L.T
                
                self.NKGg[ind,ind2.T] = L @ self.NKGg[ind,ind2.T]
                self.NKGg[ind2,ind.T] = self.NKGg[ind2,ind.T] @ L.T
                
                self.Mg[ind,ind2.T] = L @ self.Mg[ind,ind2.T]
                self.Mg[ind2,ind.T] = self.Mg[ind2,ind.T] @ L.T
            
            # Incorporación de las posibles rigideces de los resortes
            k = np.diag(np.array(N.k))
            i = self.nds.index(N)
            ind = np.array([3*i,3*i+1,3*i+2]).reshape(3,1)
            
            self.Kg[ind,ind.T] = self.Kg[ind,ind.T] + k
            

            
    def applyBC(self):
        """Se crea el vector lógico de condiciones de contorno y el vector de
        desplazamientos conocidos, ambos globales y de dimensión igual a 3 por
        el número de nodos. Se aplican durante la resolución del problema."""
        #self.BC = np.zeros((3*len(self.nds),1),dtype=np.bool)
        self.BC = []
        U = []
        
        for i in range(len(self.nds)):
            self.BC += self.nds[i].bc
            U += [self.nds[i].u[j] for j in range(3) if not self.nds[i].bc[j]]
            self.UC = np.array(U).reshape((len(U),1))
            
    def getMatrices(self):
        """Obtención de todas las matrices de la estructura, hasta las nodales 
        tras aplicación de los cambios de coordenadas necesarios"""
        self.assembleKNKGM() 
        self.getNodalMatrices()
        self.applyBC()
        
    def getFeq(self):
        """Obtención del vector de fuerzas equivalentes (Feq) y el vector de
        fuerzas concentradas (F), que se completará con el valor de las
        reacciones al resolver el problema."""
        
        tam = 3*len(self.nds)
        self.Feq = np.zeros((tam,1))
        self.F = np.zeros((tam,1))
        
        for B in self.bars:
            ind1 = self.nds.index(B.p1)
            ind2 = self.nds.index(B.p2)
            Femp = -B.T.T @ B.FempR
            
            self.Feq[3*ind1:3*ind1+3] += Femp[0:3]
            self.Feq[3*ind2:3*ind2+3] += Femp[3:6]
        
        i = 0
        for N in self.nds:
            # Conversión de Feq por apoyo no concordante
            if(N.th != 0):
                th = N.th
                L = np.mat([[cos(th),sin(th),0],[-sin(th),cos(th),0],[0,0,1]])
                
                i = self.nds.index(N)
                ind = [3*i, 3*i+1, 3*i+2]
                
                self.Feq[ind] = L @ self.Feq[ind]
                
            # Inclusión de las cargas nodales, en el sistema de coord. nodal
            if(len(N.loads) > 0):
                for L in N.loads:
                    if(L.active):
                        self.F[3*i] += L.fx
                        self.F[3*i+1] += L.fy
                        self.F[3*i+2] += L.mz
            i += 1
            
    def show(self, axes = None, bc_scale = 0.04, k_scale = 0.1, f_scale = 0.075,
             showNodes = False, node_style = defNodeStyle, bar_style = defBarStyle,
             load_style = defLoadStyle):
        # bc_scale : Escala de las condiciones de contorno        
        """Visualización de la estructura"""
        if(axes is None):
            ax = spAxes.SPAxes()
        elif(isinstance(axes, spAxes.SPAxes)):
            ax = axes
        elif(isinstance(axes, mpAxes)):
            pos = axes.get_position()
            ax = spAxes.SPAxes(axes.figure, [pos.x0,pos.y0,pos.width,pos.height])
        else:
            raise TypeError('Axes must be a matplotlib.pyplot.Axes or StruPy2D.Axes.SPAxes object.')
        
        """Cálculo del tamaño característico de la estructura."""
        x_list = self.nds.x()
        y_list = self.nds.y()
        tam = sqrt((min(y_list)-max(y_list))**2 + (min(x_list)-max(x_list))**2)
        
        """Aplicación de este factor a algunas dimensiones de los estilos"""
        load_style.dim = f_scale*tam if (load_style.dim is None) else load_style.dim
        
        """Visualización de nodos"""
        for N in self.nds:
           #ax.plot(N.x,N.y,'bs',markerfacecolor = node_style.MarkerFaceColor ,ms = node_style.MarkerSize, zorder = 1)
           node_style.visible = showNodes
           ax.addElement(gN(N, node_style))
        
        """Visualización de barras con sus libertades"""
        for B in self.bars:
            ax.addElement(gB(B, bar_style))
            
        """Visualización de las condiciones de contorno y cargas nodales"""
        for N in self.nds:
            Lbc = bc_scale * tam
            Lk = k_scale * tam
            Lf = f_scale * tam
            drawBC(ax,N,Lbc)
            
            if(True in [x!=0 for x in N.k]):
                drawElasSup(ax,N,Lk)
                
            if(True in [x != 0 for x in N.u]):
                drawDispSup(ax,N,Lbc)
            if(len(N.loads) > 0):
                #drawLoad(ax,N,Lf)
                ax.addElement(gL(N.loads[0], load_style, ax))
                                
        # https://stackoverflow.com/questions/7908636/possible-to-make-labels-appear-when-hovering-over-a-point-in-matplotlib
        """ Visualización de las cargas intraelementales """
        # visualización de las cargas intraelementales
        # Colores para secciones y/o materiales
        
        plt.axis('equal')
        plt.title(r'Estructura para cálculo a estático')
        
        if(axes == None):
            return ax
        else:
            return None
        
    def showTx(self):
        """Resumen de las propiedades de la estructura en modo texto"""